# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Handling rect in static allocation strategies
"""

from __future__ import annotations


class Rect:
    """
    Data of rectangle in static allocation strategies
    """

    def __init__(
        self, left_top: tuple[int, int], right_bottom: tuple[int, int]
    ) -> None:
        """
        :param Tuple[int,int] left_top: Left top position of the rect
        :param Tuple[int,int] right_bottom: Right bottom position of the rect
        """
        self.left_top = left_top
        self.right_bottom = right_bottom



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Heart of animation flow
Get the input and returns the output, 
    calling each modules
"""

from __future__ import annotations
from typing import Iterable
from AnimatedWordCloud.Utils import Config, TimelapseWordVector
from AnimatedWordCloud.Animator.AllocationCalculator import allocate
from AnimatedWordCloud.Animator.ImageCreator import create_images
from AnimatedWordCloud.Animator.AnimationIntegrator import integrate_images


def animate(
    word_vector_timelapse: Iterable[tuple[str, dict[str, float]]],
    config: Config = None,
    output_filename: str = "output.gif",
) -> str:
    """
    Create an animation of word cloud,
        from the timelapse data of words vectors

    This function will call each modules in the right order,
        and return the animation path

    :param Iterable[tuple[str, dict[str, float]]] word_vector_timelapse:
    Timelapse data of word vectors.
    The data structure is a list of tuples,
        which includes "name of the time(str)" and "word vector(Dict[str, float])"
    :param Config config: Configuration of the animation. If None, default config will be used.
    :param str output_filename: Filename of the animation file.
    :return: The path of the animation file.
    :rtype: str
    """

    # use default config if not specified
    if config is None:
        config = Config()

    # convert data to TimelapseWordVector
    timelapse_word_vector = TimelapseWordVector.convert_from_dicts_list(
        word_vector_timelapse
    )

    # Calculate allocation
    allocation_timelapse = allocate(timelapse_word_vector, config)

    # to images
    image_paths = create_images(allocation_timelapse, config)

    # to one animation file
    animation_path = integrate_images(
        image_paths, allocation_timelapse, config, filename=output_filename
    )

    if config.verbosity == "minor" or config.verbosity == "debug":
        _success_message(animation_path)

    return animation_path


def _success_message(animation_path: str) -> None:
    """
    print success message.

    :param str animation_path: animation saved path
    :return: None
    :rtype: None
    """
    success_message = (
        f"your animation was successfully saved in {animation_path}"
    )
    string_frame_over = (
        "-" * (len(success_message) // 2)
        + "Success!"
        + "-" * (len(success_message) // 2)
    )
    string_frame_under = "-" * len(success_message + "Success!")
    success_message = " " * 4 + success_message + " " * 4
    print(string_frame_over)
    print(success_message)
    print(string_frame_under)



from AnimatedWordCloud.Utils.Data.AllocationData import (
    AllocationInFrame,
    AllocationTimelapse,
)
from AnimatedWordCloud.Utils.Data.TimelapseWordVector import (
    WordVector,
    TimeFrame,
    TimelapseWordVector,
)
from AnimatedWordCloud.Utils.Data.Rect import Rect
from AnimatedWordCloud.Utils.Data.Word import Word

__all__ = [
    "AllocationInFrame",
    "AllocationTimelapse",
    "WordVector",
    "TimeFrame",
    "TimelapseWordVector",
    "Rect",
    "Word",
]



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Config class for passing parameters through the modules
"""

from __future__ import annotations
import random
from typing import Literal
from AnimatedWordCloud.Utils.Consts import (
    DEFAULT_ENG_FONT_PATH,
    DEFAULT_OUTPUT_PATH,
)


class Config:
    """
    Config class for AnimatedWordCloud

    :param str font_path: Path to the font file.
        If None, default English font will be used.
    :param str output_path: Path to the output directory.
        If None, default output directory in this library will be used.
        Warning: This directory won't be deleted automatically.
            So, becareful of your storage.
    :param int max_words: Maximum number of words shown in each frame.
    :param int max_font_size: Maximum font size of the word.
    :param int min_font_size: Minimum font size of the word.
    :param int image_width: Width of the image.
    :param int image_height: Height of the image.
    :param str background_color: Background color of the image, default is "white".
    :param str color_map: Colormap to be used for the image, default is "magma".
    :param str allocation_stategy: Strategy to allocate words.
        There are "magnetic" only for now.
    :param int image_division: The number of division of the image.
        This is available for magnetic strategy only.
    :param float movement_reluctance: Reluctance of the movement of the word. If higher, the word tends to stay near to the previous position.
    :param str verbosity: Verbosity of the log.
        "silent" for no log, "minor" for only important logs, "debug" for all logs.
    :param str transition_symbol: Symbol to be used for showing transition.
        It will be shown as "(former) [transition_symbol] (latter)".
    :param str starting_time_stamp: Time stamp of the starting time. (Before the first time stamp in the input data)
    :param int duration_per_interpolation_frame: Duration of each interpolation frame in milliseconds.
    :param int duration_per_static_frame: Duration of each static frame in milliseconds.
    :param int n_frames_for_interpolation: Number of frames in the animation.
    :param str interpolation_method: Method to interpolate the frames.
        There are "linear" only for now.
    :param bool drawing_time_stamp: Whether to draw time stamp on the image.
    :param str time_stamp_color: Color of the time stamp.
    :param int time_stamp_font_size: Font size of the time stamp.
        If None(default), it will be set to 75% of max_font_size
    :param tuple[int, int] time_stamp_position: Position of the time stamp.
        If None(default), it will be set to (image_width*0.75, image_height*0.75) which is right bottom.
    :param str intermediate_frames_id: Static images of each frame of itermediate product will be saved as "{intermediate_frames_id}_{frame_number}.png".
        If None(default), this will be set randomly.
    """

    def __init__(
        self,
        font_path: str = None,
        output_path: str = None,
        max_words: int = 100,
        max_font_size: int = 50,
        min_font_size: int = 10,
        image_width=800,
        image_height=600,
        background_color: str = "white",
        color_map: str = "Dark2",
        allocation_strategy: Literal["magnetic"] = "magnetic",
        image_division: int = 300,
        movement_reluctance: float = 0.05,
        verbosity: Literal["silent", "minor", "debug"] = "silent",
        transition_symbol: str = " to ",
        starting_time_stamp: str = " ",
        duration_per_interpolation_frame: int = 50,
        duration_per_static_frame: int = 700,
        n_frames_for_interpolation: int = 20,
        interpolation_method: Literal["linear"] = "linear",
        drawing_time_stamp: bool = True,
        time_stamp_color: str = "black",
        time_stamp_font_size: int = None,
        time_stamp_position: tuple[int, int] = None,
        intermediate_frames_id: str = None,
    ) -> None:
        # explanation written above

        # handle nones
        if font_path is None:
            font_path = DEFAULT_ENG_FONT_PATH
        if output_path is None:
            output_path = DEFAULT_OUTPUT_PATH
        self.font_path = font_path
        self.output_path = output_path
        self.max_words = max_words
        self.max_font_size = max_font_size
        self.min_font_size = min_font_size
        self.image_width = image_width
        self.image_height = image_height
        self.background_color = background_color
        self.color_map = color_map
        self.allocation_strategy = allocation_strategy
        self.image_division = image_division
        self.movement_reluctance = movement_reluctance
        self.verbosity = verbosity
        self.transition_symbol = transition_symbol
        self.starting_time_stamp = starting_time_stamp
        self.duration_per_interpolation_frame = (
            duration_per_interpolation_frame
        )
        self.duration_per_static_frame = duration_per_static_frame
        self.n_frames_for_interpolation = n_frames_for_interpolation
        self.interpolation_method = interpolation_method
        self.drawing_time_stamp = drawing_time_stamp
        self.time_stamp_color = time_stamp_color

        self.time_stamp_font_size = self._compute_time_stamp_font_size(
            time_stamp_font_size
        )

        self.time_stamp_position = self._compute_time_stamp_position(
            time_stamp_position
        )

        self.intermediate_frames_id = self._compute_intermediate_frames_id(
            intermediate_frames_id
        )

    def _compute_time_stamp_font_size(
        self, time_stamp_font_size: int | None
    ) -> int:
        """
        Compute time_stamp_font_size from constructor argument.
        :param int time_stamp_font_size: Font size of the time stamp by the argument.
        :return: Font size of the time stamp.
        """
        if time_stamp_font_size is None:
            return int(self.max_font_size * 0.75)
        else:
            return time_stamp_font_size

    def _compute_time_stamp_position(
        self, time_stamp_position: tuple[int, int] | None
    ) -> tuple[int, int]:
        """
        Compute time_stamp_position from constructor argument.
        :param tuple[int, int] time_stamp_position: Position of the time stamp by the argument.
        :return: Position of the time stamp.
        """
        if time_stamp_position is None:
            return (
                self.image_width * 0.75,
                self.image_height * 0.75,
            )  # right bottom
        else:
            return time_stamp_position

    def _compute_intermediate_frames_id(
        self, intermediate_frames_id: str | None
    ) -> str:
        """
        returns intermediate_frames_id given from constructor
        set random value if intermediate_frames_id is None
        :param str|None intermediate_frames_id: intermediate_frames_id given
        :return: intermediate_frames_id computed
        :rtype: str
        """

        if intermediate_frames_id is None:
            # set randomly
            return str(random.randint(0, 1000000000))
        else:
            return intermediate_frames_id



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Collision detection functions
"""

from __future__ import annotations
from typing import Iterable
from AnimatedWordCloud.Utils.Vector import Vector
from AnimatedWordCloud.Utils.Data.Rect import Rect


def is_point_hitting_rects(
    point: tuple[int, int] | Vector, rects: Iterable[Rect]
) -> tuple[bool, Rect]:
    """
    Check if the point is hitting any of the rects given.

    :param Tuple[int,int]|Vector point: Point to check
    :param Iterable[Rect] rects: Rectangles to check
    :return: (Is hitting, Rect that is hitting)
    :rtype: Tuple[bool, Rect]
    """
    for rect in rects:
        if is_point_hitting_rect(point, rect):
            return (True, rect)

    return (False, None)


def is_point_hitting_rect(point: tuple[int, int] | Vector, rect: Rect) -> bool:
    """
    Check if the point is hitting the single rect given.

    :param Tuple[int,int]|Vector point: Point to check
    :param Rect rect: Rectangle to check
    :return: Is hitting
    :rtype: bool
    """

    if point.__class__ == tuple:
        point = Vector(point)

    return (
        rect.left_top[0] < point.x < rect.right_bottom[0]
        and rect.left_top[1] < point.y < rect.right_bottom[1]
    )


def is_rect_hitting_rect(rect0: Rect, rect1: Rect) -> bool:
    """
    Check if 2 rects hitting

    Check the collision by AABB algorithm

    :param Rect rect0: Rect to check
    :param Rect rect1: Rect to check
    :return: Whether 2 rects are hitting
    :rtype: bool
    """

    return not (
        rect0.right_bottom[0] < rect1.left_top[0]
        or rect0.left_top[0] > rect1.right_bottom[0]
        or rect0.right_bottom[1] < rect1.left_top[1]
        or rect0.left_top[1] > rect1.right_bottom[1]
    )


def is_rect_hitting_rects(rect: Rect, rects: Iterable[Rect]) -> bool:
    """
    Check if the rect is hitting any of the rects given.

    :param Rect rect: Rect to check
    :param Iterable[Rect] rects: Rectangles to check
    :return: Is hitting
    :rtype: bool
    """
    for rect_to_check in rects:
        if is_rect_hitting_rect(rect, rect_to_check):
            return True

    return False



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Consts for AnimatedWordCloud functions
"""


from pathlib import Path
import os


LIBRARY_DIR = Path(__file__).parent.parent
"""
Directory of the library currently exists.
Shows "AnimatedWordCloud" directory.
"""

DEFAULT_ENG_FONT_PATH = os.path.join(
    LIBRARY_DIR,
    "Assets",
    "Fonts",
    "NotoSansMono-VariableFont_wdth,wght.ttf",
)
"""
path of default Eng font file exists
"""

DEFAULT_OUTPUT_PATH = os.path.join(LIBRARY_DIR, "output")
"""
Default output path of the generated images when none is specified.
"""



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Calculate allocation of each words in each static time
"""

from __future__ import annotations
import numpy as np
from tqdm import tqdm
from PIL import Image, ImageFont, ImageDraw
from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies import (
    MagneticAllocation,
    allocate_randomly,
)
from AnimatedWordCloud.Utils import (
    WordVector,
    TimelapseWordVector,
    AllocationInFrame,
    AllocationTimelapse,
    Word,
    Config,
)


def allocate(
    word_vector: WordVector,
    allocation_before: AllocationInFrame,
    config: Config,
) -> AllocationInFrame:
    """
    Calculate allocation of each words in each static time

    :param WordVector word_vector: The word vector
    :param AllocationInFrame allocation_before: Allocation data of the previous frame
    :param Config config: Config instance
    :return: Allocation data of the frame
    :rtype: AllocationInFrame
    """

    word_weights = word_vector.get_ranking(0, config.max_words)

    words: list[Word] = []

    # get attributes for each words,
    #   and save them as Word instances
    for word_raw, weight in word_weights:
        # get size
        font_size = calculate_font_size(
            weight,
            word_weights[0][1],  # max weight
            word_weights[-1][1],  # min weight
            config.max_font_size,
            config.min_font_size,
        )
        text_size = estimate_text_size(word_raw, font_size, config.font_path)

        # make instance
        word = Word(word_raw, weight, font_size, text_size)

        # save instance
        words.append(word)

    # calculate allocation by selected strategy
    if config.allocation_strategy == "magnetic":
        allocator = MagneticAllocation(config)
        return allocator.allocate(words, allocation_before)
    else:
        raise ValueError(
            "Unknown strategy: {}".format(config.allocation_strategy)
        )


def calculate_font_size(
    weight: float,
    weight_max: float,
    weight_min: float,
    font_max: int,
    font_min: int,
) -> int:
    """
    Evaluate how much font size the word should be

    Simply calculate linearly

    :param float weight: The weight of the word
    :param float weight_max: The maximum weight of the word
    :param float weight_min: The minimum weight of the word
    :param int font_max: The maximum font size
    :param int font_min: The minimum font size
    :return: The font size estimated
    :rtype: int
    """

    # calculate font size linearly
    weight_ratio = (weight - weight_min) / (weight_max - weight_min)
    font_size = int((font_max - font_min) * weight_ratio + font_min)

    return font_size


def estimate_text_size(
    word: str, font_size: int, font_path: str
) -> tuple[int, int]:
    """
    Estimate text box size

    Highly depends on the drawing library

    :param str word: The word
    :param int font_size: The font size
    :param str font_path: The font path
    :return: Text box size (x, y)
    :rtype: tuple[int, int]
    """

    # according to https://watlab-blog.com/2019/08/27/add-text-pixel/

    # prepare empty image
    image = np.zeros(
        (font_size * 2, font_size * (len(word) + 1), 3), dtype=np.uint8
    )
    font = ImageFont.truetype(font_path, font_size)
    image = Image.fromarray(image)
    draw = ImageDraw.Draw(image)

    # get size
    w = draw.textlength(word, font=font, font_size=font_size)
    h = font_size

    return (w, h)


def allocate_all(
    timelapse: TimelapseWordVector, config: Config
) -> AllocationTimelapse:
    """
    Calculate allocation of each words in each static time

    :param TimelapseWordVector timelapse: The timelapse word vector
    :param Config config: Config instance
    :return: Allocation data of all the frames
    :rtype: AllocationTimelapse
    """

    times = len(timelapse)

    allocation_timelapse = AllocationTimelapse()

    # first frame
    first_frame = _allocate_first_frame(timelapse[0].word_vector, config)
    allocation_timelapse.add(config.starting_time_stamp, first_frame)

    # verbose for iteration
    if config.verbosity in ["debug", "minor"]:
        print("Start static-allocation iteration...")
        iterator = tqdm(range(times))
    else:
        iterator = range(times)

    # calculate allocation for each frame
    for cnt in iterator:
        allocation = allocate(
            timelapse[cnt].word_vector,
            allocation_timelapse.get_frame(
                cnt
            ),  # first added -> cnt; one before
            config,
        )
        allocation_timelapse.add(timelapse[cnt].time_name, allocation)

    return allocation_timelapse


def _allocate_first_frame(
    word_vector: WordVector, config: Config
) -> AllocationInFrame:
    """
    Calculate allocation of the first frame

    :param WordVector word_vector: The word vector
    :param Config config: Config instance
    :return: Allocation data of the frame
    :rtype: AllocationInFrame
    """

    words_tup = word_vector.get_ranking(0, config.max_words)
    words = []  # Word instances

    # get attributes for each words,
    #   and save them as Word instances
    for word_raw, weight in words_tup:
        # minimum font size for the first frame
        text_size = estimate_text_size(
            word_raw, config.min_font_size, config.font_path
        )

        # make instance
        word = Word(word_raw, weight, config.min_font_size, text_size)

        # save instance
        words.append(word)

    # allocate randomly
    return allocate_randomly(words, config.image_width, config.image_height)



from AnimatedWordCloud.Utils.Consts import (
    LIBRARY_DIR,
    DEFAULT_ENG_FONT_PATH,
    DEFAULT_OUTPUT_PATH,
)

from AnimatedWordCloud.Utils.Config import Config

from AnimatedWordCloud.Utils.Vector import Vector

from AnimatedWordCloud.Utils.Data import (
    AllocationInFrame,
    AllocationTimelapse,
    WordVector,
    TimeFrame,
    TimelapseWordVector,
    Rect,
    Word,
)

from AnimatedWordCloud.Utils.Collisions import (
    is_point_hitting_rects,
    is_point_hitting_rect,
    is_rect_hitting_rect,
    is_rect_hitting_rects,
)

from AnimatedWordCloud.Utils.FileManager import (
    ensure_directory_exists,
)

__all__ = [
    "LIBRARY_DIR",
    "DEFAULT_ENG_FONT_PATH",
    "DEFAULT_OUTPUT_PATH",
    "Vector",
    "AllocationInFrame",
    "AllocationTimelapse",
    "WordVector",
    "TimeFrame",
    "TimelapseWordVector",
    "Rect",
    "Word",
    "is_point_hitting_rects",
    "is_point_hitting_rect",
    "is_rect_hitting_rect",
    "is_rect_hitting_rects",
    "Config",
    "ensure_directory_exists",
]



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Modules for 2D vector calculation
"""

from __future__ import annotations


class Vector:
    """
    Class for 2D vector calculation
    """

    def __init__(self, *args: float | tuple[float, float]) -> None:
        """
        Initialize Vector class.

        :param args: x and y or tuple of x and y
        input as Vector(x, y) or Vector((x, y))
        """

        # if the input is tuple...
        if args[0].__class__ == tuple:
            # ...ensuring (float, float)
            if len(args[0]) != 2:
                raise ValueError("Tuple must have 2 elements")

            # if the value was not a number, this intends to raise an error
            self.x = float(args[0][0])
            self.y = float(args[0][1])
        elif len(args) == 2:
            # ...x, y are seperatedly input

            # if the value was not a number, this intends to raise an error
            self.x = float(args[0])
            self.y = float(args[1])
        else:
            raise ValueError("Invalid Vector input")

    def __add__(self, other: Vector) -> Vector:
        """
        Sum of two vectors.

        :return: sum of two vectors
        :rtype: Vector
        """
        return Vector(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Vector) -> Vector:
        """
        Difference of two vectors.

        :return: difference of two vectors
        :rtype: Vector
        """
        return Vector(self.x - other.x, self.y - other.y)

    def __mul__(self, other: float) -> Vector:
        """
        Product of vector and scalar.

        :return: product of vector and scalar
        :rtype: Vector
        """
        return Vector(self.x * other, self.y * other)

    def __truediv__(self, other: float) -> Vector:
        """
        Division of vector and scalar.

        :return: division of vector and scalar
        :rtype: Vector
        """
        return Vector(self.x / other, self.y / other)

    def __getitem__(self, index: int) -> float:
        """
        Get item by index.

        :return: x if index is 0, y if index is 1
        :rtype: float
        """
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        else:
            raise IndexError(f"Index {index} is out of range")

    def __str__(self) -> str:
        """
        Convert to string.

        :return: string representation of the vector
        :rtype: str
        """
        return f"({self.x}, {self.y})"

    def clone(self) -> Vector:
        """
        Clone this vector.

        So as modifying the output doesn't affect the original vector.

        :return: Cloned vector
        :rtype: Vector
        """
        return Vector(self.x, self.y)

    def convert_to_tuple(self) -> tuple[float, float]:
        """
        Convert to tuple

        :return: tuple of x and y
        :rtype: tuple[float, float]
        """
        return (self.x, self.y)

    def cross(vec0: Vector, vec1: Vector) -> float:
        """
        Cross product of two vectors.
        This is static method.

        :param Vector vec0: Vector 0
        :param Vector vec1: Vector 1
        :return: Cross product of two vectors
        :rtype: float
        """
        return vec0.x * vec1.y - vec0.y * vec1.x



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Handful class of containing timelapse data of word vectors.
"""

from __future__ import annotations
from typing import Iterable
import bisect


class WordVector:
    """
    Contains each word's weight
    """

    def __init__(self) -> None:
        """
        Prepare empty data
        """

        # use bisect to easily get the rankings
        #
        # to order by weight,
        #   the weight must be the first element of the tuple
        #   and negate the weight to get the descending order
        # but the output must be the word first
        #
        # Words should be inserted by `add()`,
        #   otherwise the order will be broken
        self._word_bisect: list[tuple[float, str]] = []

        # also prepare a dictionary for direct access to word
        self._word_dictionary: dict[str, float] = {}

    def add(self, word: str, weight: float) -> None:
        """
        Add a word to the data

        :param str word: The word
        :param float weight: The weight of the word
        :rtype: None
        """

        # negate the weight to get the descending order
        bisect.insort(self._word_bisect, (-weight, word))

        self._word_dictionary[word] = weight

    def add_multiple(self, word_weights: Iterable[tuple[str, float]]) -> None:
        """
        Add multiple words to the data

        :param Iterable[Tuple[str, float]] word_weight: The words and their weights
        :rtype: None
        """

        for word, weight in word_weights:
            self.add(word, weight)

    def get_ranking(self, start: int, end: int) -> list[tuple[str, float]]:
        """
        Get the ranking of the words

        This is simply a slice of the bisect_list.

        :param int start: Start of the ranking.
        :param int end: End of the ranking. This index will not included. (such as list slice)
        If -1, to the last. If exceeds the length, to the last.
        :return: The ranking of the words
        :rtype: List[Tuple(str, float)]
        """

        # weight was negated
        #   so negate it back
        if end == -1:
            return [(tup[1], -tup[0]) for tup in self._word_bisect[start:]]
        else:
            end = min(end, len(self._word_bisect))
            return [(tup[1], -tup[0]) for tup in self._word_bisect[start:end]]

    def get_weight(self, word: str) -> float:
        """
        Get the weight of the word

        :param str word: The word
        :return: The weight of the word
        :rtype: float
        """
        return self._word_dictionary[word]

    def convert_from_dict(word_weights: dict[str, float]) -> WordVector:
        """
        Convert from a dictionary of word and weight to WordVector instance.

        This is static conversion method.

        :param Dict[str, float] word_weights: The words and their weights
        :return: The WordVector instance
        :rtype: WordVector
        """

        instance = WordVector()

        for word, weight in word_weights.items():
            instance.add(word, weight)

        return instance


class TimeFrame:
    """
    A single time frame of word vector
    """

    def __init__(self, time_name: str, word_vector: WordVector) -> None:
        """
        Prepare the time frame

        :param str time_name: Name of the time
        :param WordVector word_vector: Word vector
        """

        self.time_name: str = str(time_name)  # ensure time_name is string
        self.word_vector: WordVector = word_vector

    def convert_from_dict(
        time_name: str, word_weights: dict[str, float]
    ) -> TimeFrame:
        """
        Convert from a dictionary of word and weight to TimeFrame instance.

        This is static conversion method.

        :param str time_name: Name of the time
        :param Dict[str, float] word_weights: The words and their weights
        :return: The TimeFrame instance
        :rtype: TimeFrame
        """

        word_vector = WordVector.convert_from_dict(word_weights)

        return TimeFrame(time_name, word_vector)

    def convert_from_tup_dict(data: Iterable[str, dict[str, float]]):
        """
        Convert from a dictionary of word and weight to TimeFrame instance.

        This is static conversion method.

        :param Iterable[str, Dict[str, float]] data: The words and their weights
        :return: The TimeFrame instance
        :rtype: TimeFrame
        """

        return TimeFrame.convert_from_dict(data[0], data[1])


class TimelapseWordVector:
    """
    Timelapse data of word vectors.

    The data structure is a list of TimeFrame
    """

    def __init__(self) -> None:
        """
        Prepare empty data
        """

        # main data
        self.timeframes: list[TimeFrame] = []

    def __getitem__(
        self, index: int | list[int]
    ) -> TimeFrame | list[TimeFrame]:
        """
        Returns the item at the given index

        :param int|list[int] index: The index or slice
        :return: The timeframe at the given index
        :rtype: TimeFrame|List[TimeFrame]
        """
        return self.timeframes[index]

    def __len__(self) -> int:
        """
        Returns the length of the timelapse data

        :return: The length of the timelapse data
        :rtype: int
        """
        return len(self.timeframes)

    def add_time_frame(self, timeframe: TimeFrame) -> None:
        """
        Add a time frame to the timelapse data

        :param str time_name: Name of the time
        :param Dict[str, float] word_vector: Word vector
        :rtype: None
        """

        self.timeframes.append(timeframe)

    def convert_from_dicts_list(
        data: Iterable[Iterable[str, dict[str, float]]]
    ) -> TimelapseWordVector:
        """
        Convert from a list of dictionary of word and weight to TimelapseWordVector instance.

        This is static conversion method.

        :param Iterable[Iterable[str, Dict[str, float]]] data: list[(time_name, Dict[word, weight])]
        :return: The TimelapseWordVector instance
        :rtype: TimelapseWordVector
        """

        instance = TimelapseWordVector()

        for word_weights in data:
            instance.add_time_frame(
                TimeFrame.convert_from_tup_dict(word_weights)
            )

        return instance



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Classes that includes the data of size and position of each words
"""

from __future__ import annotations


class AllocationInFrame:
    """
    Positions and size of each words in a frame
    """

    def __init__(self, from_static_allocation: bool) -> None:
        """
        Prepare empty data

        :param bool flag_static: Whether the frame is static or not
        """

        self.from_static_allocation = from_static_allocation

        # word -> (font size, left-top position)
        self.words: dict[str, tuple[float, tuple[float, float]]] = {}

    def add(
        self, word: str, font_size: float, left_top: tuple[float, float]
    ) -> None:
        """
        Add a word

        :param str word: Word to add
        :param float font_size: Font size of the word
        :param tuple[float, float] left_top: Left-top position of the word
        :rtype: None
        """

        self.words[word] = (font_size, left_top)

    def __getitem__(self, word: str) -> tuple[float, tuple[float, float]]:
        """
        Get the data of the word

        :param str word: Word to get
        :return: (font size, left-top position)
        :rtype: tuple[float, tuple[float, float]]
        """

        return self.words[word]


class AllocationTimelapse:
    """
    Timelapse of positions and size of each words
    """

    def __init__(self) -> None:
        """
        Prepare empty data
        """

        # (time_name, AllocationInFrame)
        self.timelapse: list[tuple(str, AllocationInFrame)] = []

    def add(
        self, time_name: str, allocation_in_frame: AllocationInFrame
    ) -> None:
        """
        Add a frame of allocation data

        :param str time_name: Name of the time
        :param AllocationInFrame allocation_in_frame: Allocation data of the frame
        :rtype: None
        """

        # add
        # ensure time_name is string
        self.timelapse.append((str(time_name), allocation_in_frame))

    def get_frame(self, index: int) -> AllocationInFrame:
        """
        Get the allocation data of the frame

        :param int index: Index of the frame
        :return: Allocation data of the frame
        :rtype: AllocationInFrame
        """

        return self.timelapse[index][1]



from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies.StaticAllocationStrategy import (
    StaticAllocationStrategy,
)

from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies.MagneticAllocation import (
    MagneticAllocation,
)

from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies.RandomAllocation import (
    allocate_randomly,
)

__all__ = [
    "StaticAllocationStrategy",
    "allocate_randomly",
    "MagneticAllocation",
]



from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationCalculator import (
    allocate,
    allocate_all,
)

__all__ = [
    "allocate",
    "allocate_all",
]



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
One strategy to allocate words in static time.

First, put the largest word in the center.
Regard this word as a magnet.
Then, put the next largest word at empty point, contacting with the magnet at the center.
The point will be evaluated by evaluate_position(), and the most best point will be selected.
Repeating this process, all words will be allocated.
"""

from __future__ import annotations
import math
from typing import Iterable
from tqdm import tqdm
import joblib
from AnimatedWordCloud.Utils import (
    is_rect_hitting_rects,
    AllocationInFrame,
    Vector,
    Rect,
    Word,
    Config,
)
from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies.StaticAllocationStrategy import (
    StaticAllocationStrategy,
)
from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies.MagneticAllocation.MagnetOuterFrontier import (
    MagnetOuterFrontier,
    get_magnet_outer_frontier,
)


class MagneticAllocation(StaticAllocationStrategy):
    def __init__(
        self,
        config: Config,
    ):
        """
        Initialize allocation settings
        """
        super().__init__(config)

        self.interval_x = self.config.image_width / self.config.image_division
        self.interval_y = self.config.image_height / self.config.image_division

    def allocate(
        self, words: Iterable[Word], allocation_before: AllocationInFrame
    ) -> AllocationInFrame:
        """
        Allocate words in magnetic strategy

        :param Iterable[Word] words: Words to allocate. Order changes the result.
        :param AllocationInFrame allocation_before: Allocation data at one static frame before
        :return: Allocation data of the frame
        :rtype: AllocationInFrame
        """

        self.words = words
        self.allocations_before = allocation_before

        # missing word for previous frame
        self.add_missing_word_to_previous_frame(allocation_before, words)

        output = AllocationInFrame(from_static_allocation=True)
        magnet_outer_frontier = MagnetOuterFrontier()

        # Word rectangles that are currenly putted at the outermost of the magnet
        self.rects = set()

        # put the first word at the center
        self.center = (
            self.config.image_width / 2,
            self.config.image_height / 2,
        )
        first_word = self.words[0]
        first_word_position = (
            self.center[0] - first_word.text_size[0] / 2,
            self.center[1] - first_word.text_size[1] / 2,
        )

        # register the first word
        rect_adding = Rect(
            first_word_position,
            (
                first_word_position[0] + first_word.text_size[0],
                first_word_position[1] + first_word.text_size[1],
            ),
        )
        self.rects.add(rect_adding)
        output.add(first_word.text, first_word.font_size, first_word_position)

        # verbose for iteration
        if self.config.verbosity == "debug":
            print("MagneticAllocation: Start iteration...")
            iterator = tqdm(self.words[1:])
        else:
            iterator = self.words[1:]

        # from second word
        for word in iterator:
            # get outer frontier of the magnet
            # The position candidates will be selected from this frontier
            magnet_outer_frontier = get_magnet_outer_frontier(
                self.rects,
                self.config.image_width,
                self.config.image_height,
                self.interval_x,
                self.interval_y,
                rect_adding,
                magnet_outer_frontier,
            )

            # find the best left-top position
            position = self._find_best_position(
                word,
                magnet_outer_frontier,
                self.allocations_before[word.text][1],
            )

            # update for next iteration
            rect_adding = Rect(
                position,
                (
                    position[0] + word.text_size[0],
                    position[1] + word.text_size[1],
                ),
            )

            # register rect
            self.rects.add(rect_adding)

            # register to output
            output.add(word.text, word.font_size, position)

        # add missing words for this frame
        self.add_missing_word_from_previous_frame(allocation_before, output)

        return output

    def _evaluate_position(
        self, position_from: tuple[int, int], position_to: tuple[int, int]
    ) -> float:
        """
        Evaluate the position the word beginf to put

        :param tuple[int,int] position_from:
            Position of the center of the word comming from
        :param tuple[int,int] position_to:
            Position of the center of the word going to be putted
        :param tuple[int,int] center: Position of the center of the magnet
        :return: Evaluation value. Smaller is the better
        """

        distance_movement = math.sqrt(
            (position_from[0] - position_to[0]) ** 2
            + (position_from[1] - position_to[1]) ** 2
        )

        distance_center = math.sqrt(
            (position_to[0] - self.center[0]) ** 2
            + (position_to[1] - self.center[1]) ** 2
        )

        # the larger, the better; This need manual adjustment
        # adjust the coefficient mannually by visual testing

        # log(distance_movement): more important when near, not when far
        return (
            -self.config.movement_reluctance
            * math.log(distance_movement + 0.01)
            - 1.0 * distance_center**2
        )

    def _find_best_position(
        self,
        word: Word,
        magnet_outer_frontier: MagnetOuterFrontier,
        position_from: tuple[int, int],
    ) -> tuple[int, int]:
        """
        Find the best position to put the word

        Find the best position to put the word in the `magnet_outer_frontier`.
        The positions will be evaluated by `evaluate_position()`, and the best scored position will be returned.

        :param Word word: Word to put
        :param MagnetOuterFrontier magnet_outer_frontier:
            Outer frontier of the magnet at the center
        :param tuple[int,int] position_from:
            Position of the center of the word comming from
        :return: Best left-top position to put the word
        :rtype: tuple[int,int]
        """

        # + interval for a buffer for the collision detection
        # if not, this might get into the word's rects
        x_half = word.text_size[0] / 2 + self.interval_x
        y_half = word.text_size[1] / 2 + self.interval_y

        # Prepare for iteration
        pivots_to_center_list = [
            # from lower
            [
                Vector(0, y_half),
            ],
            # from top
            [
                Vector(0, -y_half),
            ],
            # from left
            [
                Vector(-x_half, 0),
            ],
            # from right
            [
                Vector(x_half, 0),
            ],
        ]

        frontier_sides = [
            magnet_outer_frontier.from_down,
            magnet_outer_frontier.from_up,
            magnet_outer_frontier.from_left,
            magnet_outer_frontier.from_right,
        ]

        # get center position candidates
        results_by_sides = joblib.Parallel(n_jobs=-1, verbose=0)(
            joblib.delayed(self._get_candidates_from_one_side)(
                pivots_to_center_list[cnt], frontier_sides[cnt]
            )
            for cnt in range(4)
        )
        center_position_candidates = []
        for results_by_side in results_by_sides:
            center_position_candidates.extend(results_by_side)

        # error handling: too small image area that cannot put the word anywhere anymore
        if len(center_position_candidates) == 0:
            raise Exception(
                "No available position found. Try to reduce font size or expand image size."
            )

        # find the best position
        best_position = self._try_put_all_candidates(
            center_position_candidates, word.text_size, position_from
        )

        # to left-top position
        best_position_left_top = (
            best_position[0] - word.text_size[0] / 2,
            best_position[1] - word.text_size[1] / 2,
        )

        return best_position_left_top

    def _get_candidates_from_one_side(
        self,
        pivots_to_center: Iterable[Vector],
        points_on_side: Iterable[Vector],
    ) -> list[tuple[int, int]]:
        """
        Get all candidates of the center position from one side

        Intended to be used in `find_best_position()`

        :param Iterable[Vector] pivots_to_center:
            Vector of  (center of the word) - (pivot)
        :param Iterable[Vector] points_on_side:
            Points on the side
        :return: Candidates of the center position
        :rtype: list[tuple[int, int]]
        """

        # get all candidate center positions
        candidates = []
        for point_on_side in points_on_side:
            for pivot_to_center in pivots_to_center:
                candidates.append(point_on_side + pivot_to_center)

        return candidates

    def _try_put_all_candidates(
        self,
        center_positions: Iterable[tuple[int, int]],
        size: tuple[int, int],
        position_from: tuple[int, int],
    ) -> tuple[int, int]:
        """
        Try to put the word at the gived place and evaluate the score, and return the best scored position

        :param Iterable[tuple[int,int]] center_positions:
            Candidate list of center points of the word
        :param tuple[int,int] size: Size of the word
        :param tuple[int,int] position_from:
            Position of the center of the word comming from
        :return: Best center position
        :rtype: tuple[int, int]
        """

        results_evaluation = joblib.Parallel(n_jobs=-1, verbose=0)(
            joblib.delayed(self._try_put_position)(
                center_position, size, position_from
            )
            for center_position in center_positions
        )

        # find best score
        best_position = None
        best_score = -float("inf")
        for cnt, result_evaluation in enumerate(results_evaluation):
            if (result_evaluation is not None) and (
                result_evaluation > best_score
            ):
                best_score = result_evaluation
                best_position = center_positions[cnt]

        # guard
        if best_position is None:
            raise Exception(
                "No available position found. Try to reduce font size or expand image size."
            )

        return best_position

    def _is_hitting_other_words(
        self,
        center_position: tuple[int, int] | Vector,
        size: tuple[int, int] | Vector,
    ) -> bool:
        """
        Check if the given rect is hitting with other words

        :param tuple[int,int] | Vector center_position: Center point of the word
        :param tuple[int,int] | Vector size: Size of the word
        :return: True if the center point is hitting with other words
        :rtype: bool
        """
        # ensure to Vector
        if center_position.__class__ == tuple:
            center_position = Vector(center_position)
        if size.__class__ == tuple:
            size = Vector(size)

        left_top = center_position - size / 2
        right_bottom = center_position + size / 2

        return is_rect_hitting_rects(
            Rect(
                left_top.convert_to_tuple(),
                right_bottom.convert_to_tuple(),
            ),
            self.rects,
        )

    def _try_put_position(
        self,
        center_position: tuple[float, float],
        size: tuple[float, float],
        position_from: tuple[float, float],
    ) -> float:
        """
        Evaluate the position the word if putted

        :param tuple[float,float] center_position: Position of the center of the word
        :param tuple[float,float] size: Size of the word
        :param tuple[float,float] position_from: Position of the center of the word comming from
        :return: Evaluation score. Smaller is the better
        :rtype: float
        """

        # if hitting with other word...
        if self._is_hitting_other_words(center_position, size):
            # ...skip this position
            return None

        score = self._evaluate_position(position_from, center_position)

        return score



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Calculate interpolated allocations between timestamps
"""

from __future__ import annotations
import numpy as np
from AnimatedWordCloud.Utils import (
    AllocationTimelapse,
    AllocationInFrame,
    Config,
)


def animated_allocate(
    allocation_timelapse: AllocationTimelapse, config: Config
) -> AllocationTimelapse:
    """
    :param AllocationTimelapse allocation_timelapse: static allocations already calculated (AllocationTimelapse)
    :param Config config:
    :return: AllocationTimelapse which interpolation allocation for animation inserted (AllocationTimelapse)
    :rtype: AllocationTimelapse
    """

    # final output
    timelapse_output: AllocationTimelapse = AllocationTimelapse()

    # Branching by interpolation_method
    n_timestamps = len(
        allocation_timelapse.timelapse
    )  # get the number of timestamps

    # Interpolate between timestamps. the positions of words are changed by linear.
    for index in range(n_timestamps - 1):
        # get static frame data
        from_allocation_frame = allocation_timelapse.get_frame(index)
        to_allocation_frame = allocation_timelapse.get_frame(index + 1)
        time_name_from = allocation_timelapse.timelapse[index][0]
        time_name_to = allocation_timelapse.timelapse[index + 1][0]

        # interpolate between two frames
        interpolated_frames: list[
            AllocationInFrame
        ] = _get_interpolated_frames(
            from_allocation_frame,
            to_allocation_frame,
            config,
        )

        # add static frame first
        timelapse_output.add(time_name_from, from_allocation_frame)

        # add interpolated frames
        time_name = time_name_from + config.transition_symbol + time_name_to
        for interpolated_frame in interpolated_frames:
            timelapse_output.add(time_name, interpolated_frame)

    # add last static frame
    timelapse_output.add(time_name_to, to_allocation_frame)

    return timelapse_output


def _get_setdiff(
    from_allocation_frame: AllocationInFrame,
    to_allocation_frame: AllocationInFrame,
) -> tuple[list[str], list[str]]:
    """
    Extract the keys to be added

    :param AllocationInFrame from_allocation_frame, to_allocation_frame: start frame and end frame
    :return: the keys to be added
    """
    from_words: list[str] = list(from_allocation_frame.words.keys())
    to_words: list[str] = list(to_allocation_frame.words.keys())

    # np.setdiff1(x, y)'s returns an array x excluding the elements contained in array y.
    from_words_to_be_added_key = np.setdiff1d(to_words, from_words)
    to_words_to_be_added_key = np.setdiff1d(from_words, to_words)
    return from_words_to_be_added_key, to_words_to_be_added_key


def _add_key_in_allocation_frame(
    from_allocation_frame: AllocationInFrame,
    to_allocation_frame: AllocationInFrame,
    from_words_to_be_added_key: list[str],
    to_words_to_be_added_key: list[str],
) -> tuple[AllocationInFrame, AllocationInFrame]:
    """
    :param  AllocationInFrame from_allocation_frame, to_allocation_frame: start frame and end frame
    :param list[str] from_words_to_be_added_key, to_words_to_be_added_key: the keys to be added
    :return: AllocationInFrame from_allocation_frame, to_allocation_frame: start and end frame added to the necessary keys
    :rtype: tuple[AllocationInFrame, AllocationInFrame]
    """
    for to_be_added_key in from_words_to_be_added_key:
        # add key in from_allocation_frame
        (font_size, left_top) = to_allocation_frame[to_be_added_key]
        from_allocation_frame.add(to_be_added_key, font_size, left_top)

    for to_be_added_key in to_words_to_be_added_key:
        # add key in to_allocation_frame
        (font_size, left_top) = from_allocation_frame[to_be_added_key]
        to_allocation_frame.add(to_be_added_key, font_size, left_top)

    return from_allocation_frame, to_allocation_frame


def _calc_frame_value(
    from_value: float, to_value: float, index: int, config: Config
) -> float:
    """
    :param float from_value, to_value: font_size, x_position, or y_position
    :param int index: interpolation frame index
    :param Config config used for n_frames_for_interpolation and interpolation_method
    :return: calculated interpolation's value
    :rtype: float

    Linear only for now
    Add non-Linear interpolation processes in the future.
    """
    if config.interpolation_method == "linear":
        value = _calc_linear(from_value, to_value, index, config)
    else:
        raise NotImplementedError()
    return value


def _calc_linear(
    from_value: float, to_value: float, index: int, config: Config
):
    """
    :param float from_value, to_value: font_size, x_position, or y_position
    :param int index: interpolation frame index
    :param Config config used for n_frames_for_interpolation
    :return: calculated linear interpolation's value
    :rtype: float

    Linear only for now
    Add non-Linear interpolation processes in the future.
    """
    value = from_value + index / (config.n_frames_for_interpolation + 1) * (
        to_value - from_value
    )
    return value


def _calc_added_frame(
    from_allocation_frame: AllocationInFrame,
    to_allocation_frame: AllocationInFrame,
    key: str,
    index: int,
    config: Config,
) -> tuple[float, float, float]:
    """
    :param float from_allocation_frame, to_allocation_frame: start frame and end frame
    :param str key: a focused word. It's interpolation font size and positions are calculated.
    :param int index: interpolation frame index
    :param Config config: used for the argument of _calc_frame_value
    :return: calculated interpolation's values (frame_font_size, frame_x_pos, frame_y_pos)
    :rtype: tuple[float, float, float]
    """
    from_font_size = from_allocation_frame[key][0]
    to_font_size = to_allocation_frame[key][0]
    from_x_pos = from_allocation_frame[key][1][0]
    to_x_pos = to_allocation_frame[key][1][0]
    from_y_pos = from_allocation_frame[key][1][1]
    to_y_pos = to_allocation_frame[key][1][1]
    frame_font_size = _calc_frame_value(
        from_font_size, to_font_size, index, config
    )
    frame_x_pos = _calc_frame_value(from_x_pos, to_x_pos, index, config)
    frame_y_pos = _calc_frame_value(from_y_pos, to_y_pos, index, config)
    return frame_font_size, frame_x_pos, frame_y_pos


def _get_interpolated_frames(
    from_allocation_frame: AllocationInFrame,
    to_allocation_frame: AllocationInFrame,
    config: Config,
) -> list[AllocationInFrame]:
    """
    get_interpolated_frames Algorithm:
    1. Align word counts in from_allocation_frame and to_allocation_frame
    2. Calculate interpolated frames using each of these methods for example, linear

    :param AllocationInFrame from_allocation_frame, to_allocation_frame: start frame and end frame
    :param Config config:
    :return: interpolated frames
    :rtype: list[AllocationInFrame]
    """
    # Linear only for now
    n_frames_for_interpolation = config.n_frames_for_interpolation
    from_words_to_be_added_key, to_words_to_be_added_key = _get_setdiff(
        from_allocation_frame, to_allocation_frame
    )
    from_allocation_frame, to_allocation_frame = _add_key_in_allocation_frame(
        from_allocation_frame,
        to_allocation_frame,
        from_words_to_be_added_key,
        to_words_to_be_added_key,
    )
    to_be_added_frames: list[dict[str, tuple[float, tuple[float, float]]]] = [
        {} for _ in range(n_frames_for_interpolation)
    ]  # not [{}] * n_frames_for_interpolation
    # dict[str, tuple[float, tuple[float, float]]]: word -> (font size, left-top position)
    all_keys = list(from_allocation_frame.words.keys())
    for key in all_keys:
        for index in range(1, n_frames_for_interpolation + 1):
            # from_value + index / (n_frames_for_interpolation + 1) * (to_value - from_value)
            # index: 1, 2, ..., n_frames_for_interpolation, so 1 - indexed
            frame_font_size, frame_x_pos, frame_y_pos = _calc_added_frame(
                from_allocation_frame, to_allocation_frame, key, index, config
            )
            to_be_added_frames[index - 1][key] = (
                frame_font_size,
                (frame_x_pos, frame_y_pos),
            )

    # Generate interpolated_frames as AllocationInFrame
    output = []
    for index, to_be_added_frame in enumerate(to_be_added_frames):
        to_be_added_allocation_frame = AllocationInFrame(
            from_static_allocation=False
        )
        to_be_added_allocation_frame.words = to_be_added_frame
        output.append(to_be_added_allocation_frame)

    return output



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Base class for static allocation strategies
"""

from typing import Iterable
from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies.RandomAllocation import (
    put_randomly,
)
from AnimatedWordCloud.Utils import Word, AllocationInFrame, Config


class StaticAllocationStrategy:
    """
    Base class for static allocation strategies
    """

    def __init__(self, config: Config):
        self.config = config

    def allocate(
        self, words: Iterable[Word], allocation_before: AllocationInFrame
    ) -> AllocationInFrame:
        """
        Allocate the words

        This is abstract method.

        :param Iterable[Word] words: The words to allocate
        :param AllocationInFrame allocation_before: The allocation data of the previous frame
        """

        raise NotImplementedError

    def add_missing_word_to_previous_frame(
        self, frame_previous: AllocationInFrame, words_current: Iterable[Word]
    ) -> None:
        """
        Add word in instance of the previous frame

        Find words going to be putted in current frame
            but not in previous frame.
        Add the missing words randomly put the word in the previous frame.
        The putting algorithm is the same as the one in `RandomAllocation`.
        This is intended to let appearing words appear smoothly.

        :param AllocationInFrame frame_previous:
            The allocation data of the previous frame
        :param Iterable[Word] words_current:
            The words to be putted in the current frame
        :rtype: None
        """

        # find missing words
        words_previous = set(frame_previous.words.keys())
        word_texts_current = []
        words_size = {}
        words_font_size = {}
        for word in words_current:
            if word.text not in words_current:
                word_texts_current.append(word.text)
                words_size[word.text] = word.text_size
                words_font_size[word.text] = word.font_size
        missing_words = set(word_texts_current) - words_previous

        # add missing words
        for word in missing_words:
            # put randomly
            text_lefttop_position = put_randomly(
                self.config.image_width,
                self.config.image_height,
                words_size[word],
            )

            # allocate in the output
            frame_previous.add(
                word, words_font_size[word], text_lefttop_position
            )

    def add_missing_word_from_previous_frame(
        self,
        frame_previous: AllocationInFrame,
        frame_current: AllocationInFrame,
    ) -> None:
        """
        Add missing word from the previous frame
            to the instance of the current frame

        Find words existed in previous frame but not in current,
            and add them to the current frame.
        Added words are putted randomly, same as `RandomAllocation`.
        This must be called after all words are allocated in the current frame.
        This is intended to let disappearing words disappear smoothly.

        :param AllocationInFrame frame_previous:
            The allocation data of the previous frame
        :param AllocationInFrame frame_current:
            The allocation data of the current frame
        :rtype: None
        """

        # find missing words
        words_previous = set(frame_previous.words.keys())
        words_current = set(frame_current.words.keys())
        words_missing = words_previous - words_current

        # add missing words
        for word in words_missing:
            estimated_size = (
                len(word) * frame_previous[word][0],
                frame_previous[word][0],
            )
            lefttop_position = put_randomly(
                self.config.image_width,
                self.config.image_height,
                estimated_size,
            )
            frame_current.add(word, frame_previous[word][0], lefttop_position)



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Create images of each frame
"""


from __future__ import annotations
import os
import hashlib
import numpy as np
import matplotlib.pyplot as plt
import joblib
from PIL import Image, ImageDraw, ImageFont
from AnimatedWordCloud.Utils import (
    ensure_directory_exists,
    Config,
    AllocationTimelapse,
    AllocationInFrame,
)


class colormap_color_func(object):
    # https://github.com/amueller/word_cloud/blob/main/wordcloud/wordcloud.py#L91
    """Color func created from matplotlib colormap.

    :param colormap : string or matplotlib colormap
        Colormap to sample from
    """

    def __init__(self, color_map="dark2"):
        self.color_map = plt.get_cmap(name=color_map)

    def __call__(
        self, word: str, font_size, position, random_state=None, **kwargs
    ):
        """
        To maintain each word's color
        making a word -> color surjective function
        """
        # calculate MD5 hash
        md5_hash = hashlib.md5(word.encode()).hexdigest()
        # transform hexadecimal into decimal
        decimal_value = int(md5_hash, 16)
        # Normalize from 0 to 1 value
        normalized_value = decimal_value / (16**32 - 1)
        r, g, b, _ = np.maximum(
            0, 255 * np.array(self.color_map(normalized_value))
        )
        return "rgb({:.0f}, {:.0f}, {:.0f})".format(r, g, b)


def create_image(
    allocation_in_frame: AllocationInFrame,
    config: Config,
    frame_number: int,
    time_name: str,
    color_func=None,
) -> tuple[int, str]:
    """
    Create image of a frame

    :param AllocationInFrame allocation_in_frame: Position/size data of a video frame.
    :param Config config: Config instance
    :param int frame_number: Number of the frame. Used for filename
    :param str time_name: Name of the time. Used for time stamp
    :param object color_func:  Custom function for color mapping, default is None.
    :return: (frame_number, save_path)
    :rtype: tuple[int, str]
    """
    if color_func is None:
        color_func = colormap_color_func(config.color_map)

    image = Image.new(
        "RGB",
        (config.image_width, config.image_height),
        config.background_color,
    )
    draw = ImageDraw.Draw(image)

    # Draw all words
    allocation_in_frame_word_dict = allocation_in_frame.words
    for word, position in allocation_in_frame_word_dict.items():
        font_size = position[0]
        (x, y) = position[1]
        font = ImageFont.truetype(config.font_path, font_size)
        draw.text(
            (x, y),
            word,
            fill=color_func(word=word, font_size=font_size, position=(x, y)),
            font=font,
        )

    # draw time stamp
    if config.drawing_time_stamp:
        font_size = config.time_stamp_font_size
        font = ImageFont.truetype(config.font_path, font_size)
        draw.text(
            config.time_stamp_position,
            time_name,
            fill=config.time_stamp_color,
            font=font,
        )

    # save the image
    filename = f"{config.intermediate_frames_id}_{frame_number}.png"
    save_path = os.path.join(config.output_path, filename)
    image.save(save_path)

    return (frame_number, save_path)


def create_images(
    position_in_frames: AllocationTimelapse,
    config: Config,
    color_func=None,
) -> list[str]:
    """
    Create images of each frame

    :param AllocationTimelapse position_in_frames: List of position/size data of each video frame.
    :param Config config: Config instance
    :param object color_func:  Custom function for color mapping, default is None.
    :return: The path of the images. The order of the list is the same as the order of the input.
    :rtype: list[str]
    """

    if config.verbosity in ["debug"]:
        print("Creating images of each frame...")

    ensure_directory_exists(config.output_path)

    image_paths = []

    verbosity = 0
    if config.verbosity in ["debug"]:
        verbosity = 5

    # create images of each frame
    result = joblib.Parallel(n_jobs=-1, verbose=verbosity)(
        joblib.delayed(create_image)(
            allocation_in_frame=allocation_in_frame,
            config=config,
            frame_number=frame_number,
            color_func=color_func,
            time_name=time_name,
        )
        for frame_number, (time_name, allocation_in_frame) in enumerate(
            position_in_frames.timelapse
        )
    )

    # sort by frame number (ascending)
    result.sort(key=lambda x: x[0])

    # get only the path
    image_paths = [path for _, path in result]

    return image_paths



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Data class contains attributes of each words
"""

from __future__ import annotations


class Word:
    """
    Data class contains attributes of each words.
    This is used to contact with allocation strategies.
    """

    def __init__(
        self,
        text: str,
        weight: float,
        font_size: int,
        text_size: tuple[int, int],
    ):
        self.text = text
        self.weight = weight
        self.font_size = font_size
        self.text_size = text_size



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Frontier of magnet at the center

Used by MagneticAllocation
"""


from __future__ import annotations
from typing import Iterable
from bisect import bisect
from AnimatedWordCloud.Utils import (
    Vector,
    Rect,
    is_point_hitting_rects,
    is_point_hitting_rect,
)

TO_RIGHT = Vector(0, 0)
TO_LEFT = Vector(0, 0)
TO_UP = Vector(0, 0)
TO_DOWN = Vector(0, 0)


class MagnetOuterFrontier:
    """
    Outer frontier of the magnet at the center.
    This is used to find the next position of the next word.

    This described by launching a lazer from the boarder of the image;
        from up, down, left, and right.
    And find the first point that is not overlapped with the magnet.
    """

    def __init__(self) -> None:
        """
        Make empty data
        """

        self.from_up: list[tuple[int, int]] = []
        self.from_down: list[tuple[int, int]] = []
        self.from_left: list[tuple[int, int]] = []
        self.from_right: list[tuple[int, int]] = []


def get_magnet_outer_frontier(
    rects: Iterable[Rect],
    image_width: int,
    image_height: int,
    interval_x: float,
    interval_y: float,
    rect_added: Rect,
    frontier_former: MagnetOuterFrontier,
) -> MagnetOuterFrontier:
    """
    Find the outer frontier of the magnet at the center

    :param Iterable[Rect] rects: Rectangles that are currently putted in the magnet
    :param int image_width: Width of the image
    :param int image_height: Height of the image
    :param int interval_x: interval of the precision; x
    :param int interval_y: interval of the precision; y
    :param Rect rect_added: Rectangle that is added at the last step. If specified with frontier_former, the calculation is faster.
    :param MagnetOuterFrontier frontier_former: Former frontier. If specified with rect_added, the calculation is faster.
    :return: Outer frontier of the magnet at the center
    :rtype: MagnetOuterFrontier
    """

    _initialize_directions(interval_x, interval_y)

    magnet_outer_frontier = MagnetOuterFrontier()

    # prepare for iteration
    # need to shift 1 for the collision detection
    launcher_start_positions = [
        Vector(1, 1),  # from up
        Vector(1, image_height - 1),  # from down
        Vector(1, 1),  # from left
        Vector(image_width - 1, 1),  # from right
    ]
    launcher_directions = [
        TO_RIGHT,  # from up
        TO_RIGHT,  # from down
        TO_DOWN,  # from left
        TO_DOWN,  # from right
    ]
    detection_ray_directions = [
        TO_DOWN,  # from up
        TO_UP,  # from down
        TO_RIGHT,  # from left
        TO_LEFT,  # from right
    ]
    former_frontiers_by_side = [
        frontier_former.from_up,  # from up
        frontier_former.from_down,  # from down
        frontier_former.from_left,  # from left
        frontier_former.from_right,  # from right
    ]

    # detect from 4 sides
    corresponding_frontiers = []
    for cnt in range(4):
        # detect
        detected_points = _detect_frontier_linealy(
            launcher_start_positions[cnt],
            launcher_directions[cnt],
            detection_ray_directions[cnt],
            rects,
            image_width,
            image_height,
            rect_added,
            former_frontiers_by_side[cnt],
        )

        # update
        corresponding_frontiers.append(detected_points)

    # update magnet_outer_frontier
    magnet_outer_frontier.from_up = corresponding_frontiers[0]
    magnet_outer_frontier.from_down = corresponding_frontiers[1]
    magnet_outer_frontier.from_left = corresponding_frontiers[2]
    magnet_outer_frontier.from_right = corresponding_frontiers[3]

    return magnet_outer_frontier


def _detect_frontier_linealy(
    launcher_point_start: Vector,
    launcher_direction: Vector,
    detection_ray_direction: Vector,
    rects: Iterable[Rect],
    image_width: int,
    image_height: int,
    rect_added: Rect,
    frontier_side: list[tuple[int, int]],
) -> list[tuple[int, int]]:
    """
    Detect the frontier from 1 line.

    This first set a launcher at the starting point.
    Then launch a detection ray from the launcher,
        and move the detection ray until some rect hits.
    Then move the launcher and the detection ray in the same direction,
        again and again, until the launcher is out of the image.

    :param Vector launcher_point_start:
        Starting point of the detection ray launching position
    :param Vector launcher_points_direction:
        Direction vector of the launching position moves
    :param Vector detection_ray_direction:
        Direction vector of the detection ray moves
    :param Iterable[Rect] rects:
        Word's Rectangles that are currently putted in the magnet
    :param int image_width: Width of the image
    :param int image_height: Height of the image
    :param Rect rect_added: Rectangle that is added at the last step.
    :param list[tuple[int, int]] frontier_former_side: List of points of a side if former frontier.
    :return: List of points that are detected
    :rtype: list[tuple[int, int]]
    """
    # a clone made
    frontier_side = _sort_by_direction(frontier_side, detection_ray_direction)

    # true while the launcher is in the area hitting the rect_added
    hitting = False

    image_size = (image_width, image_height)
    launcher_position = launcher_point_start.clone()

    # while lancher is inside the image...
    while is_point_hitting_rect(launcher_position, Rect((0, 0), image_size)):
        # if ray will hit the new rect...
        if _will_hit_rect_added(
            launcher_position, launcher_direction, rect_added
        ):
            # handle flag
            hitting = True

            # launch a ray
            result_ray_launched = _launch_ray(
                launcher_position,
                detection_ray_direction,
                rects,
                Rect((0, 0), image_size),
            )

            # update frontier
            _remember_ray_result(
                result_ray_launched, frontier_side, launcher_direction
            )

        # if the ray launcher escapes from the new rect hitting area...
        elif hitting:
            # ...stop
            break

        # move launcher
        launcher_position += launcher_direction

    return frontier_side


def _remember_ray_result(
    result_ray_launched: tuple[Vector, Rect] | None,
    detected_points: list[tuple[int, int]],
    launcher_direction: Vector,
) -> None:
    """
    remember the result of the ray launched.

    :param tuple[Vector, Rect]|None result_ray_launched: Result of the ray launched
    :param list[tuple[int, int]] detected_points: List of points that are detected. This will be modified.
    :param Vector launcher_direction: Direction vector of the launcher
    :rtype: None
    """

    if result_ray_launched is None:
        return

    # get the result
    detection_ray_position, _ = result_ray_launched

    # overwrite the old list
    _add_newly_found_point(
        detection_ray_position, detected_points, launcher_direction
    )


def _launch_ray(
    launching_position: Vector,
    detection_ray_direction: Vector,
    rects: Iterable[Rect],
    image_rect: Rect,
) -> tuple[Vector, Rect] | None:
    """
    Launch a detection ray from the launching position, and find the first point hits.

    This finds the frontier on the ray line. Intended to be used by `_detect_frontier_linealy()`

    :param Vector launching_position: Starting position of the ray
    :param Vector detection_ray_direction: Direction vector of the detection ray moves
    :param Iterable[Rect] rects: Rectangles that are currently putted in the magnet
    :param Rect image_rect: Rectangle of the image
    :return: If hitted -> (Position of the first point hits, Rectangle that is hitting); if not hitted -> None
    :rtype: Tuple[Vector, Rect]|None
    """
    # starting position of the ray
    # clone to avoid modifying the original vector
    detection_ray_position = launching_position.clone()

    # while detection ray is inside the image...
    while is_point_hitting_rect(detection_ray_position, image_rect):
        # check hit
        flag_hitted, hitted_rect = is_point_hitting_rects(
            detection_ray_position, rects
        )

        if flag_hitted:
            return (detection_ray_position, hitted_rect)

        # move detection ray
        detection_ray_position += detection_ray_direction

    return None


def _sort_by_direction(
    points: list[tuple[int, int]],
    direction: Vector,
) -> list[tuple[int, int]]:
    """
    Sort the points by the direction.

    :param list[tuple[int, int]] points: List of points. This won't be modified.
    :param Vector direction: Direction vector. Must be either TO_RIGHT, TO_LEFT, TO_UP, or TO_DOWN.
    :return: Sorted list of points
    :rtype: list[tuple[int, int]]
    """
    # also get components of the target direction for using bisect.bisect

    points = points.copy()

    # if direction is x-axis...
    if direction.x == 0:
        # sort by x in ascending order
        points.sort(key=lambda x: x[0])

    else:
        # sort by y in ascending order
        points.sort(key=lambda x: x[1])

    return points


def _initialize_directions(interval_x: float, interval_y: float) -> None:
    """
    Initialize the direction vectors.

    :param float interval_x: interval of the precision; x
    :param float interval_y: interval of the precision; y
    :rtype: None
    """

    global TO_RIGHT, TO_LEFT, TO_UP, TO_DOWN

    TO_RIGHT = Vector(interval_x, 0)
    TO_LEFT = Vector(-interval_x, 0)
    TO_UP = Vector(0, -interval_y)
    TO_DOWN = Vector(0, interval_y)


def _will_hit_rect_added(
    launcher_position: Vector,
    launcher_direction: Vector,
    rect_added: Rect,
) -> bool:
    """
    Check if the launcher will hit the rect_added.

    :param Vector launcher_position: Position of the launcher
    :param Vector launcher_direction: Direction vector of the launcher
    :param Rect rect_added: Rectangle that is added at the last step.
    :return: If the launcher will hit the rect_added -> True, else -> False
    :rtype: bool
    """

    # if the launcher moving vertically...
    if launcher_direction.x == 0:
        # ...check y axis
        return (
            rect_added.left_top[1]
            <= launcher_position.y
            <= rect_added.right_bottom[1]
        )

    # if the launcher moving horizontally...
    else:
        # ...check x axis
        return (
            rect_added.left_top[0]
            <= launcher_position.x
            <= rect_added.right_bottom[0]
        )


def _add_newly_found_point(
    point_found: tuple[int, int],
    frontier_points: list[tuple[int, int]],
    launcher_direction: Vector,
) -> None:
    """
    Add the newly found point to the frontier.

    :param tuple[int, int] point_found: Point found
    :param list[tuple[int, int]] frontier_points: List of points of the frontier. This will be modified.
    :param Vector launcher_direction: Direction vector of the launcher
    :rtype: None
    """

    # if the launcher moving vertically...
    if launcher_direction.x == 0:
        # find by y axis
        components = [point[1] for point in frontier_points]

        _add_newly_found_point_with_specified_component(
            components, frontier_points, point_found, point_found[1]
        )

    # if the launcher moving horizontally...
    else:
        # find by x axis
        components = [point[0] for point in frontier_points]

        _add_newly_found_point_with_specified_component(
            components, frontier_points, point_found, point_found[0]
        )


def _add_newly_found_point_with_specified_component(
    components: list[int],
    frontier_points: list[tuple[int, int]],
    point_found: tuple[int, int],
    point_component: int,
) -> None:
    """
    Update the frontier with the newly found point by _add_newly_found_point()

    :param list[int] components: List of components of the frontier points
    :param list[tuple[int, int]] frontier_points: List of points of the frontier. This will be modified.
    :param tuple[int, int] point_found: Point found
    :param int point_component: Component of the point found
    :rtype: None
    """
    index = bisect(components, point_component)

    # if there was a proceeding point...
    if (index < len(components)) and (point_component == components[index]):
        # ... overwrite
        frontier_points[index] = point_found

    # if there was no proceeding point...
    else:
        # ... newly insert

        # bisect.bisect returns the index of the point that is bigger than the target point
        frontier_points.insert(index, point_found)



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Calculate positions and size of words during each frame
"""

from AnimatedWordCloud.Utils import (
    AllocationTimelapse,
    TimelapseWordVector,
    Config,
)
from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator import (
    allocate_all as allocate_static,
)
from AnimatedWordCloud.Animator.AllocationCalculator.AnimatetdAllocationCalculator import (
    animated_allocate,
)


def allocate(
    word_vector_timelapse: TimelapseWordVector, config: Config
) -> AllocationTimelapse:
    """
    Calculate positions and size of words during each frame

    This will call
    - StaticAllocationCalculator: Calculate positions and size of words in each frame
    - AnimatedAllocationCalculator: Calculate the motion of each word in between the static frames

    :param TimelapseWordVector word_vector_timelapse:
    :return: timelapse of calculated allocation
    :rtype: AllocationTimelapse
    """

    # Calculate static allocation
    static_allocation_timelapse = allocate_static(
        word_vector_timelapse, config
    )

    # Calculate animated allocation
    animated_allocation_timelapse = animated_allocate(
        static_allocation_timelapse, config
    )

    return animated_allocation_timelapse



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Integrates the images into a single video (gif)
"""

from __future__ import annotations
import os
from PIL import Image
from AnimatedWordCloud.Utils import Config, AllocationTimelapse


def integrate_images(
    image_paths: list[str],
    allocation_timelapse: AllocationTimelapse,
    config: Config,
    filename: str = "output.gif",
) -> str:
    """
    Create images of each frame

    :param
    List[str] image_paths: List of image_paths created by AnimatedWordCloud.Animator.ImageCreator.create_images
    :param AllocationTimelapse allocation_timelapse: AllocationTimelapse instance
    :param Config config: Config instance
    :return: The path of the output animation file
    :rtype: str
    """

    if config.verbosity == "debug":
        print("Integrating images...")

    # input
    gif_images = [Image.open(path) for path in image_paths]

    # output
    filepath_output = os.path.join(config.output_path, filename)

    # compute the duration of each frame
    durations = []
    for _, allocation_in_frame in allocation_timelapse.timelapse:
        if allocation_in_frame.from_static_allocation:
            durations.append(config.duration_per_static_frame)
        else:
            durations.append(config.duration_per_interpolation_frame)

    # save gif
    gif_images[0].save(
        filepath_output,
        save_all=True,
        append_images=gif_images[1:],
        duration=durations,
        loop=0,
    )

    return filepath_output



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
First static allocation strategy.

Randomly allocate words AROUND the image.
Words are put on randomly 
    on a circle, whose center is the center of the image,
    and whose radius is 
    (the diagonal length of the image + the diagonal length of the word)/2
    which don't let the word overlap with the image in this frame.

This is used for the very first frame, before the first timelapse.
"""

from __future__ import annotations
from typing import Iterable
from random import random
from math import pi, cos, sin
from AnimatedWordCloud.Utils import Word, AllocationInFrame


def allocate_randomly(
    words: Iterable[Word], image_width: int, image_height: int
) -> AllocationInFrame:
    """
    Allocate the words

    :param Iterable[Word] words: The words to allocate
    :return: Allocation data of the frame
    :rtype: AllocationInFrame
    """

    output = AllocationInFrame(from_static_allocation=True)

    for word in words:
        # put
        text_lefttop_position = put_randomly(
            image_width, image_height, word.text_size
        )

        # allocate in the output
        output.add(word.text, word.font_size, text_lefttop_position)

    return output


def put_randomly(
    image_width: int, image_height: int, word_size: tuple[int, int]
) -> tuple[int, int]:
    """
    Randomly allocate word AROUND the image.

    Words are put on randomly
        on a circle, whose center is the center of the image,
        and whose radius is
        (the diagonal length of the image + the diagonal length of the word)/2
        which don't let the word overlap with the image in this frame.

    :param int image_width: Width of the image
    :param int image_height: Height of the image
    :param tuple[int, int] word_size: Size of the word
    :return: Left-top position of the word
    :rtype: tuple[int, int]
    """
    # calculate the radius of the circle
    image_diagnal = (image_width**2 + image_height**2) ** 0.5
    text_diagnal = (word_size[0] ** 2 + word_size[1] ** 2) ** 0.5
    radius = (image_diagnal + text_diagnal) / 2

    # randomly choose where on the circle to put
    angle = random() * 2 * pi
    image_center_x = image_width / 2
    image_center_y = image_height / 2
    x = image_center_x + radius * cos(angle)
    y = image_center_y + radius * sin(angle)
    text_lefttop_position = (
        x - word_size[0] / 2,
        y - word_size[1] / 2,
    )

    return text_lefttop_position



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Modules for controling file managements
"""

import os


def ensure_directory_exists(directory_path: str) -> None:
    """
    Verify that the directory exists. If it does not exist, create another directory.

    :param
    - str directory_path
    :rtype: None
    """
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)

    return



# -*- coding: utf-8 -*-
# Copyright (C) 2023 AnimatedWordCloud Project
# https://github.com/konbraphat51/AnimatedWordCloud
#
# Licensed under the MIT License.
"""
Setup script
"""
from setuptools import find_packages, setup
from pathlib import Path

# get README.md
readme = (Path(__file__).parent / "README.md").read_text(encoding="utf-8")


def requirements_from_file(file_name):
    return open(file_name).read().splitlines()


setup(
    name="AnimatedWordCloudTimelapse",
    version="1.0.8",
    description="Animate a timelapse of word cloud",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="konbraphat51, superhotdogcat",
    author_email="konbraphat51@gmail.com, siromisochan@gmail.com",
    url="https://github.com/konbraphat51/AnimatedWordCloud/tree/main",
    packages=find_packages(exclude=["tests", "Docs"]),
    test_suite="tests",
    python_requires=">=3.8",
    package_data={"AnimatedWordCloud": ["Assets/**"]},
    include_package_data=True,
    install_requires=[
        "Pillow==10.1.0",
        "numpy",
        "matplotlib",
        "joblib",
        "tqdm",
    ],
    license="MIT License",
    zip_safe=False,
    keywords=[
        "NLP",
        "World Cloud",
        "Animation",
        "Natural Language Processing",
        "video",
        "Visualization",
        "Data Science",
    ],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Topic :: Multimedia :: Video",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Text Processing",
    ],
    entry_points={
        # "console_scripts": [
        # ],
    },
    project_urls={
        "GitHub Repository": "https://github.com/konbraphat51/AnimatedWordCloud"
    },
)



from AnimatedWordCloud.Animator.AllocationCalculator.StaticAllocationCalculator.StaticAllocationStrategies.MagneticAllocation.MagneticAllocation import (
    MagneticAllocation,
)

__all__ = [
    "MagneticAllocation",
]
